# Copyright (c) 2015, Thomas Chiroux - Link Care Services
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.

# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.

# * Neither the name of cairotft nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
"""Transitions classes.

The Transitions are used in animations to calculates new object positions
based on some parameters like transition time, transition type, etc...

Idea and implementation is taken from mootools Fx project:
https://github.com/mootools/mootools-core/blob/master/Source/Fx/Fx.Transitions.js

see: http://www.chipwreck.de/blog/2010/02/17/mootools-transitions-explained/
"""
import math


class BaseTransition():

    """Base class for all transitions.

    This class will be subclassed by every transition type.
    """

    @classmethod
    def ease_in(cls, progress=0):
        """Method to calculate transition for easeIn display.

        Ease in is the default behaviour: Here the function is applied
        normally.
        """
        return cls.pos(progress)

    @classmethod
    def ease_out(cls, progress=0):
        """Method to calculate transition for easeOut display.

        Ease out inverts the behaviour - so the effect works backwards.
        """
        return 1 - cls.pos(1 - progress)

    @classmethod
    def ease_in_out(cls, progress=0):
        """Method to calculate transition for easeInOut display.

        Ease in and out is more complex:
            In the first half of the time the transition is applied normally,
            in the second half it is applied reverse.
        """
        if progress <= 0.5:
            return cls.pos(2 * progress) / 2
        else:
            return 2 - cls.pos(2 * (1 - progress)) / 2

    @classmethod
    def pos(cls, progress=0):
        """calculate pos based on progress value, progress is between 0 and 1.

        :param float progress: the wanted progression of the transition.
            0 is the 'beginning' of the transition, i.e. 0% of the transition
            1 is the 'end' of the transition, i.e. 100% of the transition.

        This method is defined in subclasses based on the transition type.
        """
        raise NotImplementedError


class LinearTransition(BaseTransition):

    """Simple linear transition.

    A linear transition has no additional ease-parameter,
    since it simply divides the range equally along the timeline.
    """

    @classmethod
    def pos(cls, progress=0):
        """calculate pos based on progress value, progress is between 0 and 1.

        :param float progress: the wanted progression of the transition.
        """
        return progress


class QuadTransition(BaseTransition):

    """Quad (x^2) transition.

    Polynomial (x^n) Transition.
    For these transition the value will first increase slowly and near
    the end of the effect it will grow rapidly - like the well-known parabola.
    """

    @classmethod
    def pos(cls, progress=0):
        """calculate pos based on progress value, progress is between 0 and 1.

        :param float progress: the wanted progression of the transition.
        """
        return math.pow(progress, 2)


class CubicTransition(BaseTransition):

    """Cubic (x^3) transition.

    Polynomial (x^n) Transition.
    For these transition the value will first increase slowly and near
    the end of the effect it will grow rapidly - like the well-known parabola.
    """

    @classmethod
    def pos(cls, progress=0):
        """calculate pos based on progress value, progress is between 0 and 1.

        :param float progress: the wanted progression of the transition.
        """
        return math.pow(progress, 3)


class QuartTransition(BaseTransition):

    """x^4 transition.

    Polynomial (x^n) Transition.
    For these transition the value will first increase slowly and near
    the end of the effect it will grow rapidly - like the well-known parabola.
    """

    @classmethod
    def pos(cls, progress=0):
        """calculate pos based on progress value, progress is between 0 and 1.

        :param float progress: the wanted progression of the transition.
        """
        return math.pow(progress, 4)


class QuintTransition(BaseTransition):

    """x^5 transition.

    Polynomial (x^n) Transition.
    For these transition the value will first increase slowly and near
    the end of the effect it will grow rapidly - like the well-known parabola.
    """

    @classmethod
    def pos(cls, progress=0):
        """calculate pos based on progress value, progress is between 0 and 1.

        :param float progress: the wanted progression of the transition.
        """
        return math.pow(progress, 5)


class PowTransition(BaseTransition):

    """x^6 transition.

    Polynomial (x^n) Transition.
    For these transition the value will first increase slowly and near
    the end of the effect it will grow rapidly - like the well-known parabola.
    """

    @classmethod
    def pos(cls, progress=0):
        """calculate pos based on progress value, progress is between 0 and 1.

        :param float progress: the wanted progression of the transition.
        """
        return math.pow(progress, 6)


class ExpoTransition(BaseTransition):

    """Exponancial transition.

    So the function grows even more rapidly towards the end -
    compared to the x^n-transitions.
    """

    @classmethod
    def pos(cls, progress=0):
        """calculate pos based on progress value, progress is between 0 and 1.

        :param float progress: the wanted progression of the transition.
        """
        return math.pow(2, 8 * (progress - 1))


class CircTransition(BaseTransition):

    """Circular transition.

    The circular transition function describes a perfect circle
    (in this case a quadrant)
    """

    @classmethod
    def pos(cls, progress=0):
        """calculate pos based on progress value, progress is between 0 and 1.

        :param float progress: the wanted progression of the transition.
        """
        return 1 - math.sin(math.acos(progress - 1))


class SineTransition(BaseTransition):

    """Sinus transition.

    The sinusiodal transition delivers a graph which behaves like a part
    of a sine-curve, the result in this case is a smooth, nearly linear
    transition (see the graph below).
    """

    @classmethod
    def pos(cls, progress=0):
        """calculate pos based on progress value, progress is between 0 and 1.

        :param float progress: the wanted progression of the transition.
        """
        return 1 - math.cos(progress * math.pi / 2)


class BackTransition(BaseTransition):

    """Back transition.

    Back moves the value first in the opposite direction
    and then moves it in direction of the target value.
    """

    @classmethod
    def pos(cls, progress=0):
        """calculate pos based on progress value, progress is between 0 and 1.

        :param float progress: the wanted progression of the transition.
        """
        return math.pow(progress, 2) * (2.618 * progress - 1.618)


class BounceTransition(BaseTransition):

    """Bounce transition.

    Bounce is the most complex function - it can’t be described as a
    simple formula.
    It first approaches the target value linear and then it oscillates between
    descreasing intervals, creating a bouncing effect.
    """

    @classmethod
    def pos(cls, progress=0):
        """calculate pos based on progress value, progress is between 0 and 1.

        :param float progress: the wanted progression of the transition.
        """
        value = 0
        p_a = 0
        p_b = 1
        while True:
            if progress >= (7 - 4 * p_a) / 11:
                value = p_b * p_b - math.pow(
                    (11 - 6 * p_a - 11 * progress) / 4, 2)
                break
            p_a += p_b
            p_b /= 2
        return value


class ElasticTransition(BaseTransition):

    """Elastic transition.

    Elastic oscillates in increasing intervals between the opposite direction
    and the desired one. It’s best used with ease:out or ease:in:out,
    otherwise it may happen that effect doesn’t reach its target value.
    """

    @classmethod
    def pos(cls, progress=0):
        """calculate pos based on progress value, progress is between 0 and 1.

        :param float progress: the wanted progression of the transition.
        """
        return math.pow(2, 10 * (progress - 1)) * math.cos(
            20 * progress * math.pi * 1 / 3)



# Copyright (c) 2015, Thomas Chiroux - Link Care Services
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.

# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.

# * Neither the name of cairotft nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
"""Load an display a svg image in cairo context.

This module is a modified version of the svg_image module from
WeasyPrint:
https://github.com/Kozea/WeasyPrint/blob/master/weasyprint/images.py#L99
"""
import cairosvg.parser
import cairosvg.surface


class ImageLoadingError(ValueError):

    """An error occured when loading an image.

    The image data is probably corrupted or in an invalid format.
    """

    @classmethod
    def from_exception(cls, exception):
        """Instanciate the class with the given exception."""
        name = type(exception).__name__
        value = str(exception)
        return cls('%s: %s' % (name, value) if value else name)


class ScaledSVGSurface(cairosvg.surface.SVGSurface):

    """Have the cairo Surface object have dimension in px instead of pt."""

    @property
    def device_units_per_user_units(self):
        """Force an eventual new scale.

        In this case, does nothing new.
        """
        scale = super().device_units_per_user_units
        return scale / 1


class SVGImage():

    """SVGImage class."""

    def __init__(self, base_url, svg_data=None):
        """Initialisation of the class.

        :param str base_url: path to a SVG Image.
        :param svg_data: bytestring containing the svg datas.

        You can either provide a base_url, in this case, the file will be
        loaded, or directly use svg_data with a svg content.
        """
        # Don’t pass data URIs to CairoSVG.
        # They are useless for relative URIs anyway.
        self._base_url = (
            base_url if not base_url.lower().startswith('data:') else None)
        self._svg_data = svg_data

        try:
            # cache the rendering
            self._svg = self._render()
        except Exception as exc:
            raise ImageLoadingError.from_exception(exc)
        # TODO: support SVG images with none or only one of intrinsic
        #       width, height and ratio.
        if not (self._svg.width > 0 and self._svg.height > 0):
            raise ImageLoadingError(
                'SVG images without an intrinsic size are not supported.')
        self.intrinsic_width = self._svg.width
        self.intrinsic_height = self._svg.height
        self.intrinsic_ratio = self.intrinsic_width / self.intrinsic_height

    def get_intrinsic_size(self, _image_resolution):
        """Return a tuple: (intrinsic_width, intrinsic_height)."""
        # Vector images are affected by the 'image-resolution' property.
        return self.intrinsic_width, self.intrinsic_height

    def _render(self):
        """Draw to a cairo surface but do not write to a file.

        This is a CairoSVG surface, not a cairo surface.
        """
        return ScaledSVGSurface(
            cairosvg.parser.Tree(
                bytestring=self._svg_data, url=self._base_url),
            output=None, dpi=96)

    @staticmethod
    def _scale_to_fit(image, frame, enlarge=False):
        """scale an image always keeping aspect ratio inside the frame.

        :param image: tuple of int (width, eight)
        :param fram: tuple of int (width, eight)
        :param bool enlarge: if True, do not only shrink, also enlarge.

        :returns= a (widht, eight) tuple representing the target size of the
            object.
        (thanks to jon for this method)
        """
        image_width, image_height = image
        frame_width, frame_height = frame
        image_aspect = float(image_width) / image_height
        frame_aspect = float(frame_width) / frame_height
        # Determine maximum width/height (prevent up-scaling).
        if not enlarge:
            max_width = min(frame_width, image_width)
            max_height = min(frame_height, image_height)
        else:
            max_width = frame_width
            max_height = frame_height
        # Frame is wider than image.
        if frame_aspect > image_aspect:
            height = max_height
            width = int(height * image_aspect)
        # Frame is taller than image.
        else:
            width = max_width
            height = int(width / image_aspect)
        return (width, height)

    def draw(self, context, pos_x, pos_y,
             width, height, enlarge=True, center_y=False):
        """Draw the svg image inside a box of size (width, eight).

        This draw methods keeps automatically the aspect ration of the image.

        :param context: cairo context
        :param int pos_x: x position (top left corner)
        :param int pos_y: y position (top left corner)
        :param int width: width of the box the image will fit inside
        :param int height: height of the box the image will fit inside
        :param bool enlarge: if true and if the base image is smaller than
            the box, it will enlarge the image. If False, display the original
            smaller image.
        :param bool center_y: if True add an offset to y in order to center
            the image into the box.
        """
        # svg = self._render()
        svg = self._svg  # use cached version
        image_width, image_height = self._scale_to_fit(
            image=(svg.width, svg.height),
            frame=(width, height),
            enlarge=enlarge)
        xratio = image_width / svg.width
        yratio = image_height / svg.height
        if center_y:
            y_offset = int((height - image_height) / 2)
        else:
            y_offset = 0
        context.scale(xratio, yratio)
        context.set_source_surface(svg.cairo,
                                   pos_x / xratio,
                                   (pos_y + y_offset) / yratio)
        context.paint()
        # reset the scale
        context.scale(1 / xratio, 1 / yratio)



# Copyright (c) 2015, Thomas Chiroux - Link Care Services
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.

# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.

# * Neither the name of cairotft nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
"""Class for display on tft using linuxfb."""
import asyncio
import cairocffi as cairo

from cairotft import linuxfb


class TftDisplay():

    """Display class for the tft display."""

    def __init__(self, interface='/dev/fb0', cairo_format=cairo.FORMAT_ARGB32):
        """Initialisation of the class.

        :param str interface: framebuffer interface name
        :param int cairo_format: the pixel format.
            see: https://pythonhosted.org/cairocffi/api.html#pixel-format
        """
        self.fb_interface = interface
        self.cairo_format = cairo_format

        # two memory buffers:
        #     * fbmem for direct draw on the screen
        #     * buffermem: memory buffer for double buffering.
        self.fbmem = linuxfb.open_fbmem(self.fb_interface)
        self.buffermem = linuxfb.memory_buffer(self.fbmem.fix_info.smem_len)

        # two cairo surface, directly on the screen and in the memory buffer.
        self.surf = linuxfb.cairo_surface_from_fbmem(
            self.fbmem,
            self.fbmem.mmap,
            cairo_format)
        self.buffer_surf = linuxfb.cairo_surface_from_fbmem(
            self.fbmem,
            self.buffermem,
            cairo_format)

        # calculates width and height of the screen
        self.width, self.height = self.surf.get_width(), self.surf.get_height()
        self.size_per_pixel = self.fbmem.fix_info.smem_len / (self.width *
                                                              self.height)
        # by default we write only in buffer using self.ctx
        self.ctx = cairo.Context(self.buffer_surf)

        # cairo context for direct rendering on the screen.
        # normaly only used with blit.
        self.screen_ctx = cairo.Context(self.surf)

        # async io loop
        self.io_loop = asyncio.get_event_loop()

    def blit(self):
        """Display the buffer in the screen.

        Take the content of the memory buffer and draw it on the screen.
        """
        self.screen_ctx.set_source_surface(self.buffer_surf)
        self.screen_ctx.paint()

    def close(self):
        """Close the interface."""
        # Back to black background
        self.blank_screen(self.ctx)
        linuxfb.close_fbmem(self.fbmem)

    def blank_screen(self, ctx, color=(0, 0, 0, 1), blit=True):
        """Blank the screen with the given color.

        :param ctx: cairocffi context
        :type ctx: :class:`cairocffi.Context`
        :param color: 4 int tuple reprensentig the rgba color.
        """
        ctx.set_source_rgba(*color)
        ctx.rectangle(0, 0, self.width, self.height)
        ctx.fill()
        if blit:
            self.blit()

    def draw_interface(self, ctx):
        """Method that should be overriden by subclasses."""
        raise NotImplementedError

    def run(self):
        """main loop."""
        # just afer loop is started, draw the interface
        self.io_loop.call_soon(self.draw_interface, self.ctx)
        try:
            self.io_loop.run_forever()
        except KeyboardInterrupt:
            pass
        finally:
            self.io_loop.close()
            self.close()



# Copyright (c) 2015, Thomas Chiroux - Link Care Services
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.

# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.

# * Neither the name of cairotft nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
"""BlinkIcon widget."""


class BlinkIcon():

    """A svg icon that can blink."""

    def __init__(self, display_object, svg_icon,
                 pos_x, pos_y, width, height,
                 background_color=(1, 1, 1, 1),
                 on_time=0.5, off_time=0.5):
        """Initialisation of the bliking Icon.

        :param display_object: the Display class instanciation.
        :type display_object: :class:`cairotft.yfy.TftDisplay`
        :param svg_icon: the svg icon to display.
        :type svg_icon: :class:`cairoftf.svg_image.SVGImage`
        :param int pos_x: x coordinates to display the icon
        :param int pos_y: y coordinates to display the icon
        :param int width: the width of the icon
        :param int height: the height of the icon
        :param tuple background_color: a tuple of 4 float representing the
            rgba value of the background color to repaint the icon.
        :param float on_time: the time in s the icon is displayed
        :param float off_time: the time in s the icon is not displayed

        TODO: instead of repainting with a color, save the background before
              painting the icon
        """
        self._stop = False
        self._showing = False
        self.display_object = display_object
        self.svg_icon = svg_icon
        self.pos_x = pos_x
        self.pos_y = pos_y
        self.width = width
        self.height = height
        self.background_color = background_color
        self.on_time = on_time
        self.off_time = off_time

    def show(self, ctx):
        """show the icon."""
        if not self._stop:
            self.svg_icon.draw(
                context=ctx,
                pos_x=self.pos_x,
                pos_y=self.pos_y,
                width=self.width,
                height=self.height,
                enlarge=True)
            self.display_object.blit()
            self.display_object.io_loop.call_later(
                self.on_time, self.hide, ctx)

    def hide(self, ctx):
        """hide the icon."""
        if not self._stop:
            ctx.set_source_rgba(*self.background_color)
            ctx.rectangle(self.pos_x, self.pos_y, self.width, self.height)
            ctx.fill()
            self.display_object.blit()

            self.display_object.io_loop.call_later(
                self.off_time, self.show, ctx)
        else:
            self._showing = False

    def start(self, ctx):
        """Start showing the bliking icon."""
        if not self._showing:
            self._showing = True
            self._stop = False
            self.display_object.io_loop.call_soon(
                self.show, ctx)

    def stop(self):
        """stop showing blinking icon."""
        self._stop = True
        self._showing = False



# Copyright (c) 2015, Thomas Chiroux - Link Care Services
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.

# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.

# * Neither the name of cairotft nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
"""Text Marquee widget."""
import time

import cairocffi as cairo

from cairotft import linuxfb
from cairotft import transitions


class Marquee():

    """A text marquee widget."""

    def __init__(self, display_object, text,
                 font_face, font_size,
                 text_color,
                 pos_x, pos_y, width, height,
                 background_color=(1, 1, 1, 1),
                 step=1,
                 interval_time=0.05,
                 transition=transitions.LinearTransition.ease_in,
                 smooth=False):
        """Initialisation of the Marquee.

        :param display_object: the Display class instanciation.
        :type display_object: :class:`cairotft.tft.TftDisplay`
        :param str text: the text to display.
        :param font_face: the font face used for the text
        :type font_face: :class:`cairocffi.ToyFontFace`
        :param tuple text_color: a tuple of 4 float representing the rgba
            value of the text color.
        :param int font_size: the size of the font.
        :param int pos_x: x coordinates of the text box (top left corner)
        :param int pos_y: y coordinates of the text box (top left corner)
        :param int width: the width of the text box
        :param int height: the height of the text box
        :param tuple background_color: a tuple of 4 float representing the
            rgba value of the background color to repaint the icon.
        :param int step: number of unit we skip at each frame. (X chars if
            smooth is False, X pixels if smooth is True)
        :param float interval_time: the time in s between two frames

        TODO: instead of repainting with a color, save the background or paint
              in order: from bottom to top.
        """
        self._stop = False
        self._showing = False
        self._pos = 0
        self.display_object = display_object
        self.text = text
        self.full_text = self.text + "   " + self.text
        self._shrinked_text = text
        self.font_face = font_face
        self.font_size = font_size
        self.text_color = text_color
        self.background_color = background_color
        self._old_text_color = text_color
        self._old_background_color = background_color
        self.pos_x = pos_x
        self.pos_y = pos_y
        self.width = width
        self.height = height
        self.step = step
        self.smooth = smooth
        self.max_offset = len(self.text) + 3
        self.interval_time = interval_time
        self.transition = transition
        self._transition_time = None
        self._first_time = None
        self._should_scroll = False  # if False, text does not scroll

    def _smooth_init_buffer(self, ctx):
        """Initialise the buffer used in smooth text."""
        # in smooth case, we need to instanciate specific buffers
        # at first, determine the width and height we need:
        ctx.set_font_size(self.font_size)
        ctx.set_font_face(self.font_face)
        (_, _, small_text_width, _,
         x_advance_text, _) = ctx.text_extents(self.text + "   ")
        self.smooth_text_width = int(x_advance_text)
        self.max_offset = self.smooth_text_width

        (_, y_bearing,
         width, height,
         x_advance, _) = ctx.text_extents(self.full_text)
        self.smooth_full_height = int(height)
        self.smooth_full_width = int(max(x_advance, width))
        self.smooth_pos_y = int(-y_bearing)
        stride = cairo.ImageSurface.format_stride_for_width(
            self.display_object.cairo_format, int(x_advance))
        textbuffer = linuxfb.memory_buffer(
            int(stride * self.smooth_full_height))
        self.smooth_textsurf = cairo.ImageSurface(
            format=self.display_object.cairo_format,
            width=self.smooth_full_width,
            height=self.smooth_full_height,
            data=textbuffer,
            stride=stride)
        # if the text does not fit inside the target box, we will scrool it
        if small_text_width > self.width:
            self._should_scroll = True
            self._transition_time = (self.interval_time *
                                     self.smooth_text_width / self.step)

        # creates the context for this memory surface and draw the text
        self.smooth_text_ctx = cairo.Context(self.smooth_textsurf)
        self._smooth_draw_text()

    def _smooth_draw_text(self):
        """draw the text in the smooth text buffer.

        This is not display, only drawed in the smooth buffer.
        """
        # background
        self.smooth_text_ctx.set_source_rgba(*self.background_color)
        self.smooth_text_ctx.rectangle(0,
                                       0,
                                       self.smooth_full_width,
                                       self.smooth_full_height)
        self.smooth_text_ctx.fill()
        # text
        self.smooth_text_ctx.set_font_size(self.font_size)
        self.smooth_text_ctx.set_font_face(self.font_face)
        self.smooth_text_ctx.set_source_rgba(*self.text_color)
        self.smooth_text_ctx.move_to(0, self.smooth_pos_y)
        if self._should_scroll:
            self.smooth_text_ctx.show_text(self.full_text)
        else:
            self.smooth_text_ctx.show_text(self.text)

    def _shrink_text(self, ctx):
        """based on the font size, calculate the width of the text.

        And eventually shrink the name if too long.
        """
        # same font face here than the one displayed.
        if self._pos > 0:
            new_text = (self.full_text)[self._pos:]
        else:
            new_text = self.text[self._pos:]
        while True:
            (_, _, _, _, x_advance, _) = ctx.text_extents(new_text)
            if x_advance > self.width:
                new_text = new_text[:-1]  # + '…'
            else:
                break
        self._shrinked_text = new_text
        if self._shrinked_text != self.text:
            self._should_scroll = True
            self._transition_time = (self.interval_time *
                                     len(self.text + "   ") / self.step)

    def change_color(self, color):
        """Change the text color.

        :param tuple color: a tuple of 4 float representing the
            rgba value of the text color.
        """
        self._old_text_color = self.text_color
        self.text_color = color
        if self.smooth and self._showing:
            self._smooth_draw_text()

    def change_background(self, background_color):
        """Change the background color.

        :param tuple background_color: a tuple of 4 float representing the
            rgba value of the background color to repaint the icon.
        """
        self._old_background_color = self.background_color
        self.background_color = background_color
        if self.smooth and self._showing:
            self._smooth_draw_text()

    def color_changed(self):
        """Return True is either text_color or background_color has changed.

        (since last paint)
        """
        if (self._old_text_color != self.text_color or
                self._old_background_color != self.background_color):
            return True

    def show(self, ctx, no_loop=False):
        """Show the text."""
        if not self._stop and self._showing:
            # here at each frame, move the text and display it.
            if (self.color_changed or self._should_scroll) and not self.smooth:
                # erase the text box
                ctx.set_source_rgba(*self.background_color)
                ctx.rectangle(self.pos_x, self.pos_y, self.width, self.height)
                ctx.fill()

                # display text
                ctx.set_source_rgba(*self.text_color)
                ctx.set_font_size(self.font_size)
                ctx.set_font_face(self.font_face)

                self._shrink_text(ctx)
                ctx.move_to(
                    self.pos_x,
                    (self.pos_y +
                     (self.height - self.font_size) / 2 +
                     self.font_size) - 2)
                ctx.show_text(self._shrinked_text)
                self.display_object.blit()
            elif (self.color_changed or self._should_scroll) and self.smooth:
                # erase the text box
                ctx.set_source_rgba(*self.background_color)
                ctx.rectangle(self.pos_x, self.pos_y, self.width, self.height)
                ctx.fill()
                ctx.set_source_surface(
                    self.smooth_textsurf.create_for_rectangle(
                        self._pos, 0, self.width, self.smooth_full_height),
                    self.pos_x,
                    (self.pos_y +
                     (self.height - self.smooth_full_height) / 2) + 1)
                ctx.paint()
                self.display_object.blit()

            if self._should_scroll:  # only cycle when text is too long.
                now = time.time()
                if self._first_time is not None:
                    transition_offset = self.transition(
                        (now - self._first_time) /
                        self._transition_time)
                else:
                    transition_offset = self.transition(0)
                    self._first_time = now
                self._pos = int(self.max_offset * transition_offset)
                if (now - self._first_time) > self._transition_time:
                    self._first_time = now  # recycle
                    self._pos = int(self.max_offset * self.transition(0))

            self._old_text_color = self.text_color
            self._old_background_color = self.background_color

            if not no_loop:
                self.display_object.io_loop.call_later(
                    self.interval_time, self.show, ctx)

    def start(self, ctx):
        """Start showing the marquee."""
        if not self._showing:
            self._showing = True
            self._stop = False
            self._first_time = None
            if self.smooth:
                self._smooth_init_buffer(ctx)
            else:
                self._shrink_text(ctx)

            self.display_object.io_loop.call_soon(
                self.show, ctx)

    def stop(self):
        """stop showing the marquee."""
        self._stop = True
        self._showing = False
        self._first_time = None



# Copyright (c) 2015, Thomas Chiroux - Link Care Services
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.

# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.

# * Neither the name of cairotft nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
"""setuptools installer for cairotft."""

import os
import uuid

from pip.req import parse_requirements
from setuptools import find_packages
from setuptools import setup
from setuptools.command.build_py import build_py

# local imports
try:
    from build_scripts.version import get_git_version
except:
    pass

here = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(here, 'README.rst')).read()
NEWS = open(os.path.join(here, 'NEWS.rst')).read()

version = None
try:
    version = get_git_version()
except:
    pass

if version is None or not version:
    try:
        file_name = "cairotft/RELEASE-VERSION"
        version_file = open(file_name, "r")
        try:
            version = version_file.readlines()[0]
            version = version.strip()
        except Exception:
            version = "0.0.0"
        finally:
            version_file.close()
    except IOError:
        version = "0.0.0"


class my_build_py(build_py):
    def run(self):
        # honor the --dry-run flag
        if not self.dry_run:
            target_dirs = []
            target_dirs.append(os.path.join(self.build_lib,
                                            'cairotft'))
            target_dirs.append('cairotft')
            # mkpath is a distutils helper to create directories
            for dir in target_dirs:
                self.mkpath(dir)

            try:
                for dir in target_dirs:
                    fobj = open(os.path.join(dir, 'RELEASE-VERSION'), 'w')
                    fobj.write(version)
                    fobj.close()
            except:
                pass

        # distutils uses old-style classes, so no super()
        build_py.run(self)

dev_reqs_gen = parse_requirements("dev-requirements.txt",
                                  session=uuid.uuid1())
reqs_gen = parse_requirements("requirements.txt",
                              session=uuid.uuid1())

setup(name='cairotft',
      version=version,
      description="UI library for small tft screens using cairocffi",
      long_description=README + '\n\n' + NEWS,
      cmdclass={'build_py': my_build_py},
      classifiers=[
          "Development Status :: 2 - Pre-Alpha",
          "Intended Audience :: Developers",
          "License :: OSI Approved :: BSD License",
          "Operating System :: POSIX :: Linux",
          "Programming Language :: Python",
          "Programming Language :: Python :: 3.4",
          "Topic :: Multimedia :: Graphics",
          "Topic :: Software Development :: Libraries :: Python Modules",
          "Topic :: Software Development :: User Interfaces", ],
      keywords='cairo framebuffer tft python',
      author='Thomas Chiroux - Link Care Services',
      author_email='tchiroux (at) linkcareservices.org',
      url='https://github.com/LinkCareServices/cairotft',
      license='BSD 3-Clause',
      packages=find_packages(exclude=['ez_setup']),
      package_data={'': ['*.rst', ], },
      include_package_data=True,
      zip_safe=False,
      test_suite='nose.collector',
      tests_require=[str(ir.req) for ir in dev_reqs_gen],
      install_requires=[str(ir.req) for ir in reqs_gen], )



# Copyright (c) 2012 Kurichan
#
# This program is free software. It comes without any warranty, to
# the extent permitted by applicable law. You can redistribute it
# and/or modify it under the terms of the Do What The Fuck You Want
# To Public License, Version 2, as published by Sam Hocevar. See
# http://sam.zoy.org/wtfpl/COPYING for more details.
"""Small wrapping of linux/fb.h.

update for cairotft: this module has been modified to support
        double buffering and is not api-compatible anymore with the original
        module from Kurichan.

        It has also been modified for python 3.4 compat'
"""
import os
import mmap
import ctypes
from fcntl import ioctl

#
# Working without ctypes example
#
# >>> buf = array.array('B', (0x00,)*(16+8+4))
# >>> fcntl.ioctl(fid, 0x4602, buf)
# 0
# >>> struct.unpack('I', buf[-4:])
# (2457600,)
#

FBIOGET_VSCREENINFO = 0x4600
FBIOGET_FSCREENINFO = 0x4602


class FbFid(int):

    """The framebuffer file descriptor.

    The name attribute gives the
    opened framebuffer file.
    """

    # __slots__ = ('name',)
    pass


class FixScreenInfo(ctypes.Structure):

    """The fb_fix_screeninfo from fb.h."""

    _fields_ = [
        ('id_name', ctypes.c_char * 16),
        ('smem_start', ctypes.c_ulong),
        ('smem_len', ctypes.c_uint32),
        ('type', ctypes.c_uint32),
        ('type_aux', ctypes.c_uint32),
        ('visual', ctypes.c_uint32),
        ('xpanstep', ctypes.c_uint16),
        ('ypanstep', ctypes.c_uint16),
        ('ywrapstep', ctypes.c_uint16),
        ('line_length', ctypes.c_uint32),
        ('mmio_start', ctypes.c_ulong),
        ('mmio_len', ctypes.c_uint32),
        ('accel', ctypes.c_uint32),
        ('reserved', ctypes.c_uint16 * 3),
    ]


class FbBitField(ctypes.Structure):

    """The fb_bitfield struct from fb.h."""

    _fields_ = [
        ('offset', ctypes.c_uint32),
        ('length', ctypes.c_uint32),
        ('msb_right', ctypes.c_uint32),
    ]


class VarScreenInfo(ctypes.Structure):

    """The fb_var_screeninfo struct from fb.h."""

    _fields_ = [
        ('xres', ctypes.c_uint32),
        ('yres', ctypes.c_uint32),
        ('xres_virtual', ctypes.c_uint32),
        ('yres_virtual', ctypes.c_uint32),
        ('xoffset', ctypes.c_uint32),
        ('yoffset', ctypes.c_uint32),

        ('bits_per_pixel', ctypes.c_uint32),
        ('grayscale', ctypes.c_uint32),

        ('red', FbBitField),
        ('green', FbBitField),
        ('blue', FbBitField),
        ('transp', FbBitField),
    ]


def open_fbdev(fbdev=None):
    """Return the framebuffer file descriptor.

    Try to use the FRAMEBUFFER
    environment variable if fbdev is not given. Use '/dev/fb0' by
    default.
    """
    dev = fbdev or os.getenv('FRAMEBUFFER', '/dev/fb0')
    fbfid = FbFid(os.open(dev, os.O_RDWR))
    fbfid.name = dev
    return fbfid


def close_fbdev(fbfid):
    """Close the framebuffer file descriptor."""
    os.close(fbfid)


def get_fix_info(fbfid):
    """Return the fix screen info from the framebuffer file descriptor."""
    fix_info = FixScreenInfo()
    ioctl(fbfid, FBIOGET_FSCREENINFO, fix_info)
    return fix_info


def get_var_info(fbfid):
    """Return the var screen info from the framebuffer file descriptor."""
    var_info = VarScreenInfo()
    ioctl(fbfid, FBIOGET_VSCREENINFO, var_info)
    return var_info


def map_fb_memory(fbfid, fix_info):
    """Map the framebuffer memory."""
    return mmap.mmap(
        fbfid,
        fix_info.smem_len,
        mmap.MAP_SHARED,
        mmap.PROT_READ | mmap.PROT_WRITE,
        offset=0
    )


class FbMem(object):

    """The framebuffer memory object.

    Made of:
        - the framebuffer file descriptor
        - the fix screen info struct
        - the var screen info struct
        - the mapped memory
    """

    __slots__ = ('fid', 'fix_info', 'var_info', 'mmap')


def open_fbmem(fbdev=None):
    """Create the FbMem framebuffer memory object."""
    fid = open_fbdev(fbdev)
    fix_info = get_fix_info(fid)
    fbmmap = map_fb_memory(fid, fix_info)
    fbmem = FbMem()
    fbmem.fid = fid
    fbmem.fix_info = fix_info
    fbmem.var_info = get_var_info(fid)
    fbmem.mmap = fbmmap
    return fbmem


def memory_buffer(buffer_len):
    """Create a memory buffer of buffer_len size.

    this memory buffer can be used to create a custom cairo surface for
    double buffering (or n-buffering)

    :param int buffer_len: size of the buffer.

    :return: the created buffer
    """
    _buffer = ctypes.create_string_buffer(buffer_len)
    return _buffer


def close_fbmem(fbmem):
    """Close the FbMem framebuffer memory object."""
    fbmem.mmap.close()
    close_fbdev(fbmem.fid)
    # TODO: free the double_buffer ?


def cairo_surface_from_fbmem(fbmem, mem, cairo_format):
    """Create a cairo surface from FbMem object.

    :param fbmem: framebuffer memory object
    :param mem: the memory buffer, either created directly or via mmap
    :param int cairo_format: cairo pixel format.
    """
    import cairocffi as cairo
    return cairo.ImageSurface.create_for_data(
        mem,
        cairo_format,  # cairo.FORMAT_RGB24,  # cairo.FORMAT_RGB16_565, etc...
        fbmem.var_info.xres,
        fbmem.var_info.yres,
        fbmem.fix_info.line_length)
